from __future__ import annotations

from http.client import HTTPException, HTTPResponse
import io
import os
import logging
from pathlib import Path
import shlex
from socket import socket
from urllib.parse import urljoin
from typing import ClassVar

from .meta import DEFAULT_LOGGER
from .task import Task
from .util import truncate

__all__ = [
    "Fetch",
    "FetchCurlError",
    "FetchException",
    "FetchMeter",
    "FetchParseException",
    "FetchResponse",
]


class FetchException(HTTPException):
    """Exception from Fetch."""

    pass


class FetchParseException(FetchException):
    """HTTP parse exception from Fetch."""

    pass


class FetchCurlError(FetchException):
    """Subprocess (cURL) error from Fetch."""

    def __init__(self, code: int, stderr: bytes):
        if stderr:
            msg = stderr.strip().decode("utf-8", errors="ignore")
        else:
            msg = f"curl ({code})"
        self.code = code
        super().__init__(msg)


class FetchMeter:
    """
    An abstract class that can be used with Fetch to provide a progress meter.
    """

    def sleep_duration(self) -> float | int:
        """Return the time in seconds to sleep per iteration."""
        return 0.1

    def begin(self) -> None:
        """Called at the beginning of the loop."""
        pass

    def step(self, index: int) -> None:
        """
        Called with each loop iteration.

        The index argument will provide a running count, starting at 0.
        """
        pass

    def end(self) -> None:
        """Called at the end of the loop."""
        pass


class FetchResponse(HTTPResponse):
    """Extend HTTPResponse to support fixed bytes instead of a stream."""

    def __init__(
        self,
        raw: bytes,
        url: str,
        debuglevel: int = 0,
        method: str | None = None,
    ):
        """
        The HTTPResponse initializer patched to support bytes instead of a
        socket argument.
        """

        class FakeSocket(socket):
            def __init__(self, fp: io.IOBase) -> None:
                self.fp = fp

            def makefile(self, *args, **kwargs) -> io.IOBase:  # type: ignore
                return self.fp

        raw = self._patch_raw(raw)
        raw_stream = io.BytesIO(raw)
        sock = FakeSocket(raw_stream)
        super().__init__(sock, debuglevel=debuglevel, method=method, url=url)

        self.url = url  # assign url due to a bug in HTTPResponse
        self.begin()

    def begin(self) -> None:
        """Raise a FetchException when unable to parse the raw response."""
        try:
            super().begin()
        except HTTPException:
            raise FetchParseException("cannot parse response")

    @classmethod
    def make_list_from_raw(cls, raw_list: bytes, url: str) -> list[FetchResponse]:
        """
        Return a list of FetchResponses from the raw bytes of one or more HEAD
        requests.
        """
        responses = []
        for raw in raw_list.strip().split(b"\r\n\r\n"):
            response = FetchResponse(raw=raw, url=url)
            responses.append(response)
            location = response.headers["location"]
            if location:
                url = urljoin(url, location)
        return responses

    @staticmethod
    def _patch_raw(raw: bytes) -> bytes:
        """
        Replace HTTP/x with HTTP/1.1 in a raw response so HTTPResponse.begin()
        doesn't throw an exception. Also, add a fake HTTP status line if we
        don't find one to handle non-HTTP protocols.
        """
        raw_lines = raw.strip().split(b"\r\n")
        parts = raw_lines[0].split(None, 1) if raw_lines else []
        if parts and parts[0].startswith(b"HTTP"):
            raw_lines[0] = b" ".join([b"HTTP/1.1"] + parts[1:])
        else:
            raw_lines.insert(0, b"HTTP/1.1 200 OK")
        return b"\r\n".join(raw_lines + [])


class Fetch:
    """Head / download request handler using cURL as a backend."""

    DEFAULT_CURL_PATH: ClassVar[str] = "curl"

    curl_path: Path
    logger: logging.Logger
    meter: FetchMeter | None

    def __init__(
        self,
        curl_path: os.PathLike | str | None = None,
        logger: logging.Logger | None = None,
        meter: FetchMeter | None = None,
    ) -> None:
        """
        Create a new Fetch object.

        If meter is specified, then use that during requests.
        """
        if curl_path is None:
            curl_path = self.DEFAULT_CURL_PATH
        if logger is None:
            logger = DEFAULT_LOGGER

        self.curl_path = Path(curl_path)
        self.logger = logger
        self.meter = meter

    def download(
        self,
        url: str,
        path: os.PathLike | str,
        resume: bool = False,
    ) -> None:
        """
        Download url to path using cURL.

        Raise a FetchCurlError when cURL exits with an error.
        """
        args = [
            "--silent",
            "--show-error",
            "--fail",
            "--location",
            "--output",
            str(path),
            *(["--continue-at", "-"] if resume else []),
            "--",
            url,
        ]
        self._curl_run(args)

    def head(self, url: str) -> list[FetchResponse]:
        """
        Make HEAD requests to url using cURL and return a list of
        FetchResponses.

        Raise a FetchCurlError when cURL exits with an error.
        """
        args = [
            "--silent",
            "--show-error",
            "--location",
            "--head",
            "--",
            url,
        ]
        task = self._curl_run(args)
        if isinstance(task.stdout, bytes):
            return FetchResponse.make_list_from_raw(raw_list=task.stdout, url=url)
        else:  # pragma: no cover
            raise RuntimeError(f"expected bytes task.stdout: {task.stdout!r}")

    def run(self, args: list) -> Task:
        """Run a subprocess of args and return the completed Task."""

        cmd = " ".join(shlex.quote(arg) for arg in args)
        self.logger.debug(f">>> {cmd}")

        task = Task(args)
        if self.meter:
            self.meter.begin()
            for index in task.loop(self.meter.sleep_duration()):
                self.meter.step(index)
            self.meter.end()
        task.wait()

        msg = f"<<< code={task.code} out={task.stdout!r} err={task.stderr!r}"
        self.logger.debug(truncate(msg, 120))
        return task

    # Private

    def _curl_run(self, args: list) -> Task:
        """
        Run a cURL subprocess of args and return the completed Task.

        Raise a FetchCurlError when cURL exits with an error.
        """
        args = [str(self.curl_path), *args]
        task = self.run(args)
        if task.code == 0:
            return task
        elif not isinstance(task.code, int):  # pragma: no cover
            raise RuntimeError(f"expected int task.code: {task.code!r}")
        elif not isinstance(task.stderr, bytes):  # pragma: no cover
            raise RuntimeError(f"expected bytes task.stderr: {task.stderr!r}")
        else:
            raise FetchCurlError(code=task.code, stderr=task.stderr)
